@@ -18,6 +18,9 @@ from .backrooms_error import BackroomsError
 from .rooms import Rooms
 
 INCLUDE_FILE_EXTENSION = ".brs"
+VALID_ROW_CHARACTERS = set(string.ascii_letters
+                           + string.digits
+                           + "`~-_=+!@#$%^&*()_+[{]}\\|;:'\",<.>/? ")
 
 
 class TranslatorError(BackroomsError):
@@ -31,6 +34,16 @@ class TranslatorError(BackroomsError):
         """
         return cls(f"{repr(include)} is a missing include!")
 
+    @classmethod
+    def must_include(cls,
+                     include: str) -> 'TranslatorError':
+        """
+        info: Used to indicate that a script has all ready been included.
+        :param include: str
+        :return: TranslatorError
+        """
+        return cls(f"{repr(include)} has all ready been included!")
+
     @classmethod
     def name_collision(cls, name: str):
         return cls(f"Name collision with {name}!")
@@ -53,6 +66,10 @@ class TranslatorError(BackroomsError):
             return cls(f"Bad line at {handler_name}:{line_number}: {repr(line)}!")
         return cls(f"Bad line at {handler_name}:{line_number}:{str(error)}: {repr(line)}!")
 
+    @classmethod
+    def file_not_found_error(cls, name: str):
+        return cls(f"File not found: {repr(name)}")
+
 
 class Handler:
     def __init__(self,
@@ -117,7 +134,10 @@ class FileHandler(Handler):
 
     def __next__(self) -> str:
         if self._file_iter is None:
-            self._file_handler = open(self._path)
+            try:
+                self._file_handler = open(self._path)
+            except FileNotFoundError:
+                raise TranslatorError.file_not_found_error(self._path)
             self._file_iter = iter(self._file_handler)
         return next(self._file_iter).rstrip("\n")
 
@@ -202,7 +222,7 @@ class Handlers:
             return self._line_number
 
     def include(self,
-                name: str) -> None:
+                name: str) -> bool:
         """
         info: Includes name if it has not been included all ready.
         :param name: str
@@ -216,9 +236,10 @@ class Handlers:
                 if name in name_space:
                     self._used_names.add(name)
                     self._handlers.append(name_space[name])
-                    return
+                    return True
             # could not find the include handler
             raise TranslatorError.missing_include(name)
+        return False
 
 
 def _tokenize_line(line: str) -> List[str]:
@@ -321,6 +342,8 @@ def translator(handlers: Handlers) -> Rooms:
             # row
             if line.startswith("/"):
                 # write row into rooms
+                if any((characters not in VALID_ROW_CHARACTERS for characters in line[1:])):
+                    raise TranslatorError.bad_line(full_line, handlers.get_line_number(), handlers.get_name())
                 rooms.write_line(x, y, floor, line[1:], 1, 0, 0)
                 # go down a row
                 y += -1
@@ -369,6 +392,20 @@ def translator(handlers: Handlers) -> Rooms:
                         raise TranslatorError.bad_line(full_line, handlers.get_line_number(), handlers.get_name())
                 else:
                     raise TranslatorError.bad_line(full_line, handlers.get_line_number(), handlers.get_name())
+            # must include
+            elif line.startswith("!"):
+                tokens = _tokenize_line(line[1:])
+                # check that we have include name
+                if tokens:
+                    script = tokens.pop(0)
+                    # include script and check that it was not all ready included
+                    if not handlers.include(script):
+                        raise TranslatorError.must_include(script)
+                    # check that we are out of tokens
+                    if tokens:
+                        raise TranslatorError.bad_line(full_line, handlers.get_line_number(), handlers.get_name())
+                else:
+                    raise TranslatorError.bad_line(full_line, handlers.get_line_number(), handlers.get_name())
             # parallel
             elif line.startswith("="):
                 tokens = _tokenize_line(line[1:])
